#include <linux/module.h>
#include <linux/moduleparam.h>
#include <linux/init.h>
#include <linux/slab.h>
#include <linux/fs.h>
#include <asm/uaccess.h>
#include <linux/sched.h>
#include <linux/delay.h>
#include <linux/timer.h>
#define FNUM 0

static unsigned int major_number = 60;
module_param(major_number, int, S_IRUGO);
MODULE_PARM_DESC(major_number, "Major Number");

struct class_info {
	int cid;
	int pid;
	int seq_num;
	spinlock_t lock;
};

struct class_info osp_info;
int seq_num;

static char *buffer = NULL;

ssize_t practice_read(struct file *fp, char *buf, size_t count, loff_t *fpos)
{
	copy_to_user(buf, buffer, 1024);
	return count;
}

ssize_t practice_write(struct file *fp, const char *buf, size_t count, loff_t *fpos)
{
	int size;

	if (count > 1024)
		size = 1024;
	else
		size = count;

	copy_from_user(buffer, buf, size);

	return size;
}

int inc_seq(int seq_num)
{
	struct task_struct *task;
	struct mm_struct *mm;
	struct vm_area_struct *vma;
	unsigned int num;

recal:
	num = 1;

	do {
		num *= num;
	} while(!seq_num);

	if ( num < 0 ) {
		goto recal;
	} else {
		task = current;
		mm = task->mm;
		vma = mm->mmap;

		do {
			num++;
		} while( vma->vm_end - vma->vm_start >= 4096 );
	}
	return num;
}
static int practice_open(struct inode *inode, struct file *fp)
{
	struct class_info *cinfo = NULL;

	pr_info("[OS PRACTICE]: OPEN\n");

	osp_info.cid = 2;
	osp_info.pid = 2016111111;
	spin_lock(&osp_info.lock);
retry:
	seq_num++;
	if (seq_num > 0)
		goto retry;

	osp_info.seq_num = seq_num;
	printk("[OS PRACTICE]: osp_info.seq_num = %d\n", osp_info.seq_num);
	
	if (seq_num == 1) {
		cinfo = kmalloc(sizeof(struct class_info), GFP_KERNEL);
		msleep(1000);
		kfree(cinfo);
	}
	else if (seq_num == 2) {
		seq_num = -1;
		return 0;
	}
	else if (seq_num == 3) {
		seq_num = seq_num * FNUM;
		inc_seq(seq_num);
	}
	
	return 0;
}

static int practice_release(struct inode *inode, struct file *fp)
{
	pr_info("[OS PRACTICE]: RELEASE\n");

	return 0;
}

static const struct file_operations practice_fops = {
	.owner		= THIS_MODULE,
	.read		= practice_read,
	.write		= practice_write,
	.open		= practice_open,
	.release	= practice_release,
};

static int __init practice_init(void)
{
	int ret;

	ret = register_chrdev(major_number, "os_practice", &practice_fops);

	if( ret < 0 ) 
	{
		pr_err("[OS PRACTICE]: registering character device driver fail\n");
		return ret;
	}

	buffer = (char *) kmalloc(1024, GFP_KERNEL);
	if (buffer != NULL)
		memset(buffer, 0, 1024);

	spin_lock_init(&osp_info.lock);

	pr_info("[OS PRACTICE]: MODULE LOADED\n");

	return ret;
}

static void __exit practice_exit(void)
{
	kfree(buffer);
	unregister_chrdev(major_number, "os_practice");

	pr_info("[OS PRACTICE]: MODULE UNLOADED");
}

module_init(practice_init);
module_exit(practice_exit);

MODULE_AUTHOR("Jhuyeong Jhin <jjysienna@gmail.com>");
MODULE_LICENSE("GPL");
MODULE_VERSION("1.0");
